#  Copyright (C) 2021-2026 pytest-qgis Contributors.
#
#
#  This file is part of pytest-qgis.
#
#  pytest-qgis is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  pytest-qgis is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with pytest-qgis.  If not, see <https://www.gnu.org/licenses/>.

import contextlib
import logging
import os.path
import shutil
import sys
import time
import warnings
from collections import namedtuple
from pathlib import Path
from typing import TYPE_CHECKING
from unittest import mock

import pytest
from qgis.core import (
    Qgis,
    QgsApplication,
    QgsProject,
    QgsRectangle,
    QgsSettings,
    QgsVectorLayer,
)
from qgis.gui import QgisInterface as QgisInterfaceOrig
from qgis.gui import QgsGui, QgsLayerTreeMapCanvasBridge, QgsMapCanvas
from qgis.PyQt import QtCore, QtWidgets, sip
from qgis.PyQt.QtCore import QCoreApplication, QSettings, Qt
from qgis.PyQt.QtWidgets import QMainWindow, QMessageBox, QWidget

from pytest_qgis.mock_qgis_classes import MockMessageBar
from pytest_qgis.qgis_bot import QgisBot
from pytest_qgis.qgis_interface import QgisInterface
from pytest_qgis.utils import (
    ensure_qgis_layer_fixtures_are_cleaned,
    get_common_extent_from_all_layers,
    get_layers_with_different_crs,
    replace_layers_with_reprojected_clones,
    set_map_crs_based_on_layers,
)

if TYPE_CHECKING:
    from _pytest.config import Config
    from _pytest.config.argparsing import Parser
    from _pytest.fixtures import SubRequest
    from _pytest.mark import Mark

QGIS_V4_INT = 40000

LOGGER = logging.getLogger("QGIS")

Settings = namedtuple(
    "Settings",
    [
        "gui_enabled",
        "qgis_init_disabled",
        "qgis_exit_disabled",
        "qgis_server",
        "canvas_width",
        "canvas_height",
    ],
)
ShowMapSettings = namedtuple(
    "ShowMapSettings", ["timeout", "add_basemap", "zoom_to_common_extent", "extent"]
)

GUI_DISABLE_KEY = "qgis_disable_gui"
GUI_ENABLED_KEY = "qgis_gui_enabled"
GUI_DESCRIPTION = "Set whether the graphical user interface is wanted or not."
GUI_ENABLED_DEFAULT = True

EXIT_DISABLED_KEY = "qgis_disable_exit"
EXIT_DISABLED_DEFAULT = False
EXIT_DISABLED_DESCRIPTION = "Set wether exitQgis() is called at the end of the session"

SERVER_KEY = "qgis_server"
SERVER_DEFAULT = False

CANVAS_HEIGHT_KEY = "qgis_canvas_height"
CANVAS_WIDTH_KEY = "qgis_canvas_width"
CANVAS_DESCRIPTION = "Set canvas height and width."
CANVAS_SIZE_DEFAULT = (600, 600)

DISABLE_QGIS_INIT_KEY = "qgis_disable_init"
DISABLE_QGIS_INIT_DESCRIPTION = "Prevent QGIS (QgsApplication) from initializing."

SHOW_MAP_MARKER = "qgis_show_map"
SHOW_MAP_VISIBILITY_TIMEOUT_DEFAULT = 30
SHOW_MAP_MARKER_DESCRIPTION = (
    f"{SHOW_MAP_MARKER}(timeout={SHOW_MAP_VISIBILITY_TIMEOUT_DEFAULT}, add_basemap=False, zoom_to_common_extent=True, extent=None): "  # noqa: E501
    f"Show QGIS map for a short amount of time. The first keyword, *timeout*, is the "
    f"timeout in seconds until the map closes. The second keyword *add_basemap*, "
    f"when set to True, adds Natural Earth countries layer as the basemap for the map. "
    f"The third keyword *zoom_to_common_extent*, when set to True, centers the map "
    f"around all layers in the project. Alternatively the fourth keyword *extent* "
    f"can be provided as QgsRectangle."
)

_APP: QgsApplication | None = None
_CANVAS: QgsMapCanvas | None = None
_IFACE: QgisInterface | None = None
_PARENT: QtWidgets.QWidget | None = None
_AUTOUSE_QGIS: bool | None = None

_QGIS_SERVER: bool = False

try:
    _QGIS_VERSION = Qgis.versionInt()
except AttributeError:
    _QGIS_VERSION = Qgis.QGIS_VERSION_INT


@pytest.hookimpl()
def pytest_addoption(parser: "Parser") -> None:
    group = parser.getgroup(
        "qgis",
        "Utilities for testing QGIS plugins",
    )
    group.addoption(f"--{GUI_DISABLE_KEY}", action="store_true", help=GUI_DESCRIPTION)
    group.addoption(
        f"--{DISABLE_QGIS_INIT_KEY}",
        action="store_true",
        help=DISABLE_QGIS_INIT_DESCRIPTION,
    )
    group.addoption(
        f"--{EXIT_DISABLED_KEY}",
        action="store_true",
        help=EXIT_DISABLED_DESCRIPTION,
    )

    parser.addini(
        GUI_ENABLED_KEY, GUI_DESCRIPTION, type="bool", default=GUI_ENABLED_DEFAULT
    )
    parser.addini(
        EXIT_DISABLED_KEY,
        EXIT_DISABLED_DESCRIPTION,
        type="bool",
        default=EXIT_DISABLED_DEFAULT,
    )
    parser.addini(
        SERVER_KEY,
        "QGIS Server session",
        type="bool",
        default=SERVER_DEFAULT,
    )

    parser.addini(
        CANVAS_WIDTH_KEY,
        CANVAS_DESCRIPTION,
        type="string",
        default=CANVAS_SIZE_DEFAULT[0],
    )
    parser.addini(
        CANVAS_HEIGHT_KEY,
        CANVAS_DESCRIPTION,
        type="string",
        default=CANVAS_SIZE_DEFAULT[1],
    )


@pytest.hookimpl(tryfirst=True)
def pytest_configure(config: "Config") -> None:
    """Configure and initialize qgis session for all tests."""
    config.addinivalue_line("markers", SHOW_MAP_MARKER_DESCRIPTION)

    settings = _parse_settings(config)
    config._plugin_settings = settings

    if not settings.gui_enabled:
        os.environ["QT_QPA_PLATFORM"] = "offscreen"

    _start_and_configure_qgis_app(config)


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_teardown(item: pytest.Item, nextitem: pytest.Item | None) -> None:  # noqa: ARG001
    request = item.funcargs.get("request")
    if request:
        ensure_qgis_layer_fixtures_are_cleaned(request)


@pytest.fixture(autouse=True, scope="session")
def qgis_app(request: "SubRequest") -> QgsApplication:
    yield _APP if not request.config._plugin_settings.qgis_init_disabled else None

    if not request.config._plugin_settings.qgis_init_disabled:
        assert _APP

        if not request.config._plugin_settings.qgis_server:
            # TODO: investigate why legendLayersAdded is sometimes not connected
            with contextlib.suppress(TypeError):
                QgsProject.instance().legendLayersAdded.disconnect(_APP.processEvents)
            if not sip.isdeleted(_CANVAS) and _CANVAS is not None:
                _CANVAS.deleteLater()

        if not request.config._plugin_settings.qgis_exit_disabled:
            LOGGER.debug("EXITING QGIS")
            _APP.exitQgis()


@pytest.fixture(scope="session")
def qgis_parent(qgis_app: QgsApplication) -> QWidget:  # noqa: ARG001
    return _PARENT


@pytest.fixture(scope="session")
def qgis_canvas() -> QgsMapCanvas:
    assert _CANVAS
    return _CANVAS


@pytest.fixture(scope="session")
def qgis_version() -> int:
    """QGIS version number as integer."""
    return _QGIS_VERSION


@pytest.fixture(scope="session")
def qgis_iface() -> QgisInterfaceOrig:
    # This is needed because qgis_iface
    # is required in autouse=True fixture qgis_show_map
    if not _QGIS_SERVER:
        assert _IFACE
    return _IFACE


@pytest.fixture(scope="session")
def qgis_processing(qgis_app: QgsApplication) -> None:
    """
    Initializes QGIS processing framework
    """
    _initialize_processing(qgis_app)


@pytest.fixture
def qgis_new_project(qgis_iface: QgisInterface, request: "SubRequest") -> QgsProject:
    """
    Initializes new QGIS project by removing layers and relations etc.

    :return: QgsProject instance
    """
    qgis_iface.newProject()  # noqa: QGS201

    # Clear the project properly if qgis_show_map marker is not used
    show_map_marker = request.node.get_closest_marker(SHOW_MAP_MARKER)
    if not show_map_marker:
        QgsProject.instance().clear()
    return QgsProject.instance()


@pytest.fixture
def qgis_world_map_geopackage(tmp_path: Path) -> Path:
    """
    Path to natural world map geopackage containing Natural Earth data.
    This geopackage can be modified in any way.

    Layers:
    * countries
    * disputed_borders
    * states_provinces
    """
    return _get_world_map_geopackage(tmp_path)


@pytest.fixture
def qgis_countries_layer(qgis_world_map_geopackage: Path) -> QgsVectorLayer:
    """
    Natural Earth countries as a QgsVectorLayer.
    """
    return _get_countries_layer(qgis_world_map_geopackage)


@pytest.fixture(scope="session")
def qgis_bot(qgis_iface: QgisInterface) -> QgisBot:
    """
    Object that holds common utility methods for interacting with QGIS.
    """
    return QgisBot(qgis_iface)


@pytest.fixture(autouse=True)
def qgis_show_map(
    qgis_app: QgsApplication,
    qgis_iface: QgisInterface | None,
    qgis_parent: QWidget | None,
    tmp_path: Path,
    request: "SubRequest",
) -> None:
    """
    Shows QGIS map if qgis_show_map marker is used.
    """
    # Noop if server session
    if _QGIS_SERVER:
        yield
        return

    assert qgis_iface is not None

    show_map_marker = request.node.get_closest_marker(SHOW_MAP_MARKER)
    common_settings: Settings = request.config._plugin_settings

    if show_map_marker:
        # Assign the bridge to have correct layer order and visibilities
        bridge = QgsLayerTreeMapCanvasBridge(  # noqa: F841, this needs to be assigned
            QgsProject.instance().layerTreeRoot(), qgis_iface.mapCanvas()
        )
        _show_qgis_dlg(common_settings, qgis_parent)

    yield

    if (
        show_map_marker
        and common_settings.gui_enabled
        and not common_settings.qgis_init_disabled
    ):
        _configure_qgis_map(
            qgis_app,
            qgis_iface,
            qgis_parent,
            _parse_show_map_marker(show_map_marker),
            tmp_path,
        )


def _load_qgis_settings(config: "Config") -> None:
    rootdir = config.rootpath
    path = rootdir.joinpath(".qgis-settings")

    os.environ["QGIS_CUSTOM_CONFIG_PATH"] = str(path)
    os.environ["QGIS_OPTIONS_PATH"] = str(path)

    qgis_ini_file = "QGIS3.ini" if _QGIS_VERSION < QGIS_V4_INT else "QGIS4.ini"

    settings_path = path.joinpath("profiles", "default")
    # Copy the ini file at correct location
    settings_file = settings_path.joinpath(
        QgsApplication.QGIS_ORGANIZATION_NAME,
        qgis_ini_file,
    )
    settings_file.parent.mkdir(parents=True, exist_ok=True)

    # Copy the ini file
    settings = rootdir.joinpath("qgis_settings.ini")
    if settings.exists():
        shutil.copyfile(settings, settings_file)

    if _QGIS_VERSION < QGIS_V4_INT:
        settings_format = QSettings.IniFormat
        settings_scope = QSettings.UserScope
    else:
        settings_format = QSettings.Format.IniFormat
        settings_scope = QSettings.Scope.UserScope

    QSettings.setDefaultFormat(settings_format)
    QSettings.setPath(settings_format, settings_scope, str(settings_path))

    qgssettings = QgsSettings()
    LOGGER.info("QGIS Settings loaded from %s", qgssettings.fileName())


def _start_and_configure_qgis_app(config: "Config") -> None:
    global _APP, _CANVAS, _IFACE, _PARENT  # noqa: PLW0603
    settings: Settings = config._plugin_settings

    # From qgis server
    # Will enable us to read qgis setting file
    QCoreApplication.setOrganizationName(QgsApplication.QGIS_ORGANIZATION_NAME)
    QCoreApplication.setOrganizationDomain(QgsApplication.QGIS_ORGANIZATION_DOMAIN)
    QCoreApplication.setApplicationName(QgsApplication.QGIS_APPLICATION_NAME)

    QCoreApplication.setAttribute(
        Qt.ApplicationAttribute.AA_ShareOpenGLContexts,
        True,
    )

    _load_qgis_settings(config)

    platform = "server" if settings.qgis_server else None

    if not settings.qgis_init_disabled:
        _APP = QgsApplication(
            [],
            GUIenabled=settings.gui_enabled,
            platformName=platform,
        )
        # Do not initialize QGIS app in qgis server mode
        # Since this is done in QgsServer() initialization
        if not settings.qgis_server:
            _APP.initQgis()
            QgsGui.editorWidgetRegistry().initEditors()

    if not settings.qgis_server:
        _PARENT = QMainWindow()
        _CANVAS = QgsMapCanvas(_PARENT)
        _PARENT.resize(QtCore.QSize(settings.canvas_width, settings.canvas_height))
        _CANVAS.resize(QtCore.QSize(settings.canvas_width, settings.canvas_height))

        # QgisInterface is a stub implementation of the QGIS plugin interface
        _IFACE = QgisInterface(_CANVAS, MockMessageBar(), _PARENT)

        # Patching imported iface (evaluated as None in tests) with iface
        mock.patch("qgis.utils.iface", _IFACE).start()

        if _APP is not None:
            # QGIS zooms to the layer's extent if it
            # is the first layer added to the map.
            # If the qgis_show_map marker is used, this zooming might occur
            # at some later time when events are processed (e.g. at qtbot.wait call)
            # and this might change the extent unexpectedly.
            # It is better to process events right after adding the
            # layer to avoid these kind of problems.
            QgsProject.instance().legendLayersAdded.connect(_APP.processEvents)

    if _APP is not None:
        # Initialize plugin_path is all cases
        _init_qgis_plugins_path(_APP)


def _init_qgis_plugins_path(qgis_app: QgsApplication) -> None:
    # Give access to python QGIS plugins
    python_plugins_path = os.path.join(qgis_app.pkgDataPath(), "python", "plugins")
    if python_plugins_path not in sys.path:
        LOGGER.info("QGIS plugins path: %s", python_plugins_path)
        sys.path.append(python_plugins_path)


def _initialize_processing(_qgis_app: QgsApplication) -> None:
    # Keep the unused arguments explicit because in will require
    # QgsApplication to be initialized instead of assuming it
    # implicitely.
    from processing.core.Processing import Processing  # noqa: PLC0415

    Processing.initialize()


def _show_qgis_dlg(common_settings: Settings, qgis_parent: QWidget) -> None:
    if not common_settings.qgis_init_disabled:
        qgis_parent.setWindowTitle("Test QGIS dialog opened by Pytest-qgis")
        qgis_parent.show()
    elif common_settings.qgis_init_disabled:
        warnings.warn(
            "Cannot show QGIS map because QGIS is not initialized. "
            "Run the tests without --qgis_disable_init to enable QGIS map.",
            stacklevel=1,
        )
    if not common_settings.gui_enabled:
        warnings.warn(
            "QGIS map is not visible because the GUI is not enabled. "
            "Set qgis_gui_enabled=True in pytest.ini to see the window.",
            stacklevel=1,
        )


def _configure_qgis_map(
    qgis_app: QgsApplication,
    qgis_iface: QgisInterface,
    qgis_parent: QWidget,
    settings: ShowMapSettings,
    tmp_path: Path,
) -> None:
    if settings.timeout == 0:
        qgis_parent.close()
        return

    message_box = QMessageBox(qgis_parent)

    try:
        # Change project CRS to most common CRS if it is not set
        if not QgsProject.instance().crs().isValid():
            set_map_crs_based_on_layers()

        extent = settings.extent
        if settings.zoom_to_common_extent and extent is None:
            extent = get_common_extent_from_all_layers()
        if extent is not None:
            qgis_iface.mapCanvas().setExtent(extent)

        # Replace layers with different CRS
        layers_with_different_crs = get_layers_with_different_crs()
        if layers_with_different_crs:
            _initialize_processing(qgis_app)
            replace_layers_with_reprojected_clones(layers_with_different_crs, tmp_path)

        if settings.add_basemap:
            # Add Natural Earth Countries
            countries_layer = _get_countries_layer(_get_world_map_geopackage(tmp_path))
            if not QgsProject.instance().addMapLayer(countries_layer):
                raise AssertionError(
                    f"Failed to add countries layer: {countries_layer.name()}"
                )
            if countries_layer.crs() != QgsProject.instance().crs():
                _initialize_processing(qgis_app)
                replace_layers_with_reprojected_clones([countries_layer], tmp_path)

        QgsProject.instance().reloadAllLayers()
        qgis_iface.mapCanvas().refreshAllLayers()

        message_box.setWindowTitle("pytest-qgis")
        message_box.setText(
            "Click close to close the map and to end the test.\n"
            f"It will close automatically in {settings.timeout} seconds."
        )
        message_box.addButton(QMessageBox.StandardButton.Close)
        message_box.move(QgsApplication.instance().primaryScreen().geometry().topLeft())
        message_box.setWindowModality(QtCore.Qt.WindowModality.NonModal)
        message_box.show()

        t = time.time()
        while time.time() - t < settings.timeout and message_box.isVisible():
            QCoreApplication.processEvents()
    finally:
        message_box.close()
        qgis_parent.close()


def _parse_settings(config: "Config") -> Settings:
    global _QGIS_SERVER  #  noqa: PLW0603

    qgis_server = config.getini(SERVER_KEY)
    if qgis_server:
        gui_enabled = False
        _QGIS_SERVER = True
    else:
        gui_disabled = config.getoption(GUI_DISABLE_KEY)
        if not gui_disabled:
            gui_enabled = config.getini(GUI_ENABLED_KEY)
        else:
            gui_enabled = not gui_disabled

    qgis_init_disabled = config.getoption(DISABLE_QGIS_INIT_KEY)

    qgis_exit_disabled = config.getoption(EXIT_DISABLED_KEY)
    if not qgis_exit_disabled:
        qgis_exit_disabled = config.getini(EXIT_DISABLED_KEY)

    canvas_width = int(config.getini(CANVAS_WIDTH_KEY))
    canvas_height = int(config.getini(CANVAS_HEIGHT_KEY))

    return Settings(
        gui_enabled,
        qgis_init_disabled,
        qgis_exit_disabled,
        qgis_server,
        canvas_width,
        canvas_height,
    )


def _parse_show_map_marker(marker: "Mark") -> ShowMapSettings:  # noqa: C901, PLR0912 TODO: Fix complexity
    timeout = add_basemap = zoom_to_common_extent = extent = notset = object()

    for kwarg, value in marker.kwargs.items():
        if kwarg == "timeout":
            timeout = value
        elif kwarg == "add_basemap":
            add_basemap = value
        elif kwarg == "zoom_to_common_extent":
            zoom_to_common_extent = value
        elif kwarg == "extent":
            extent = value
        else:
            raise TypeError(
                f"Invalid keyword argument for qgis_show_map marker: {kwarg}"
            )

    if len(marker.args) >= 1 and timeout is not notset:
        raise TypeError("Multiple values for timeout argument of qgis_show_map marker")
    elif len(marker.args) >= 1:
        timeout = marker.args[0]
    if len(marker.args) >= 2 and add_basemap is not notset:  # noqa: PLR2004
        raise TypeError(
            "Multiple values for add_basemap argument of qgis_show_map marker"
        )
    elif len(marker.args) >= 2:  # noqa: PLR2004
        add_basemap = marker.args[1]
    if len(marker.args) >= 3 and zoom_to_common_extent is not notset:  # noqa: PLR2004
        raise TypeError(
            "Multiple values for zoom_to_common_extent argument of qgis_show_map marker"
        )
    elif len(marker.args) >= 3:  # noqa: PLR2004
        zoom_to_common_extent = marker.args[2]
    if len(marker.args) >= 4 and extent is not notset:  # noqa: PLR2004
        raise TypeError("Multiple values for extent argument of qgis_show_map marker")
    elif len(marker.args) >= 4:  # noqa: PLR2004
        extent = marker.args[3]
    if len(marker.args) > 4:  # noqa: PLR2004
        raise TypeError("Too many arguments for qgis_show_map marker")

    if timeout is notset:
        timeout = SHOW_MAP_VISIBILITY_TIMEOUT_DEFAULT
    if add_basemap is notset:
        add_basemap = False
    if zoom_to_common_extent is notset:
        zoom_to_common_extent = True
    if extent is notset:
        extent = None
    elif not isinstance(extent, QgsRectangle):
        raise TypeError("Extent has to be of type QgsRectangle")
    return ShowMapSettings(timeout, add_basemap, zoom_to_common_extent, extent)


def _get_world_map_geopackage(tmp_path: Path) -> Path:
    """Copy geopackage to the temporary directory and return the copy."""
    world_map_gpkg = Path(
        QgsApplication.pkgDataPath(), "resources", "data", "world_map.gpkg"
    )
    assert world_map_gpkg.exists(), world_map_gpkg

    # Copy the geopackage to allow modifications
    return Path(shutil.copy(world_map_gpkg, tmp_path))


def _get_countries_layer(geopackage: Path) -> QgsVectorLayer:
    countries_layer = QgsVectorLayer(
        f"{geopackage}|layername=countries",
        "Natural Earth Countries",
        "ogr",
    )
    assert countries_layer.isValid(), geopackage
    return countries_layer



#  Copyright (C) 2021-2026 pytest-qgis Contributors.
#
#
#  This file is part of pytest-qgis.
#
#  pytest-qgis is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  pytest-qgis is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with pytest-qgis.  If not, see <https://www.gnu.org/licenses/>.

"""QGIS plugin implementation.

.. note:: This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

.. note:: This source code was copied from the 'postgis viewer' application
     with original authors:
     Copyright (c) 2010 by Ivan Mincik, ivan.mincik@gista.sk
     Copyright (c) 2011 German Carrillo, geotux_tuxman@linuxmail.org
     Copyright (c) 2014 Tim Sutton, tim@linfiniti.com

"""

__author__ = "tim@linfiniti.com"
__revision__ = "$Format:%H$"
__date__ = "10/01/2011"
__copyright__ = (
    "Copyright (c) 2010 by Ivan Mincik, ivan.mincik@gista.sk and "
    "Copyright (c) 2011 German Carrillo, geotux_tuxman@linuxmail.org"
    "Copyright (c) 2014 Tim Sutton, tim@linfiniti.com"
    "Copyright (c) 2021-2026 pytest-qgis Contributors"
)

import logging
import typing
from unittest.mock import MagicMock

from qgis.core import (
    QgsLayerTree,
    QgsMapLayer,
    QgsProject,
    QgsRasterLayer,
    QgsRelationManager,
    QgsVectorLayer,
)
from qgis.gui import QgsMapCanvas
from qgis.PyQt import sip
from qgis.PyQt.QtCore import QObject, QSize, pyqtSignal, pyqtSlot
from qgis.PyQt.QtWidgets import (
    QAction,
    QDockWidget,
    QMainWindow,
    QMenuBar,
    QToolBar,
    QWidget,
)

from pytest_qgis.mock_qgis_classes import MockMessageBar

LOGGER = logging.getLogger("QGIS")


# noinspection PyMethodMayBeStatic,PyPep8Naming
class QgisInterface(QObject):
    """Class to expose QGIS objects and functions to plugins.

    This class is here for enabling us to run unit tests only,
    so most methods are simply stubs.
    """

    currentLayerChanged = pyqtSignal(QgsMapCanvas)  # noqa: N815
    newProjectCreated = pyqtSignal()  # noqa: N815

    def __init__(
        self, canvas: QgsMapCanvas, messageBar: MockMessageBar, mainWindow: QMainWindow
    ) -> None:
        """Constructor
        :param canvas:
        """
        QObject.__init__(self)
        self.canvas = canvas
        self._messageBar = messageBar
        self._mainWindow = mainWindow
        self._active_layer_id: str | None = None

        # Set up slots so we can mimic the behaviour of QGIS when layers
        # are added.
        LOGGER.debug("Initialising canvas...")
        # noinspection PyArgumentList
        QgsProject.instance().layersAdded.connect(self.addLayers)
        # noinspection PyArgumentList
        QgsProject.instance().removeAll.connect(self.removeAllLayers)

        # For processing module
        self.destCrs = None
        self._layers: list[QgsMapLayer] = []

        # Add the MenuBar
        menu_bar = QMenuBar()
        self._mainWindow.setMenuBar(menu_bar)

        # Add the toolbar list
        self._toolbars: dict[str, QToolBar] = {}

        # Store mocks for missing methods to ensure consistency
        self._mock_methods: dict[str, MagicMock] = {}

    def __getattr__(self, name: str) -> MagicMock:
        """
        Return a MagicMock for any method not explicitly implemented.
        """
        if name not in self._mock_methods:
            self._mock_methods[name] = MagicMock(name=name)
        return self._mock_methods[name]

    @pyqtSlot("QList<QgsMapLayer*>")
    def addLayers(self, layers: list[QgsMapLayer]) -> None:
        """Handle layers being added to the registry so they show up in canvas.

        :param layers: list<QgsMapLayer> list of map layers that were added

        .. note:: The QgsInterface api does not include this method,
            it is added here as a helper to facilitate testing.
        """
        # LOGGER.debug('addLayers called on qgis_interface')
        # LOGGER.debug('Number of layers being added: %s' % len(layers))
        # LOGGER.debug('Layer Count Before: %s' % len(self.canvas.layers()))
        current_layers = self.canvas.layers()
        final_layers = []
        for layer in current_layers:
            final_layers.append(layer)
        for layer in layers:
            final_layers.append(layer)
        self._layers = final_layers

        self.canvas.setLayers(final_layers)
        # LOGGER.debug('Layer Count After: %s' % len(self.canvas.layers()))

    @pyqtSlot()
    def removeAllLayers(self) -> None:
        """Remove layers from the canvas before they get deleted."""
        if not sip.isdeleted(self.canvas):
            self.canvas.setLayers([])
        self._layers = []

    def newProject(self, promptToSaveFlag: bool = False) -> bool:  # noqa: ARG002
        """Create new project."""
        # noinspection PyArgumentList
        instance = QgsProject.instance()
        instance.removeAllMapLayers()
        root: QgsLayerTree = instance.layerTreeRoot()
        root.removeAllChildren()
        relation_manager: QgsRelationManager = instance.relationManager()
        for relation in relation_manager.relations():
            relation_manager.removeRelation(relation)
        self._layers.clear()
        self._mock_methods.clear()
        self._messageBar.clear_messages()
        self.newProjectCreated.emit()
        return True

    # ---------------- API Mock for QgsInterface follows -------------------

    def zoomFull(self) -> None:
        """Zoom to the map full extent."""

    def zoomToPrevious(self) -> None:
        """Zoom to previous view extent."""

    def zoomToNext(self) -> None:
        """Zoom to next view extent."""

    def zoomToActiveLayer(self) -> None:
        """Zoom to extent of active layer."""

    def addVectorLayer(
        self, path: str, base_name: str, provider_key: str
    ) -> QgsVectorLayer:
        """Add a vector layer.

        :param path: Path to layer.
        :type path: str

        :param base_name: Base name for layer.
        :type base_name: str

        :param provider_key: Provider key e.g. 'ogr'
        :type provider_key: str
        """
        layer = QgsVectorLayer(path, base_name, provider_key)
        self.addLayers([layer])
        return layer

    @typing.overload
    def addRasterLayer(
        self, rasterLayerPath: str | None, baseName: str | None = None
    ) -> QgsRasterLayer | None:
        pass

    @typing.overload
    def addRasterLayer(
        self, url: str | None, layerName: str | None, providerKey: str | None
    ) -> QgsRasterLayer | None:
        pass

    def addRasterLayer(
        self, *args: str, **kwargs: dict[str, str]
    ) -> QgsRasterLayer | None:
        layer = QgsRasterLayer(*args, **kwargs)
        self.addLayers([layer])
        return layer

    def activeLayer(self) -> QgsMapLayer | None:
        """Get pointer to the active layer (layer selected in the legend)."""
        return (
            QgsProject.instance().mapLayer(self._active_layer_id)
            if self._active_layer_id
            else None
        )

    def addPluginToMenu(self, name: str, action: QAction) -> None:
        """Add plugin item to menu.

        :param name: Name of the menu item
        :type name: str

        :param action: Action to add to menu.
        :type action: QAction
        """
        pass

    def addToolBarIcon(self, action: QAction) -> None:
        """Add an icon to the plugins toolbar.

        :param action: Action to add to the toolbar.
        :type action: QAction
        """
        pass

    def removeToolBarIcon(self, action: QAction) -> None:
        """Remove an action (icon) from the plugin toolbar.

        :param action: Action to add to the toolbar.
        :type action: QAction
        """
        pass

    def addToolBar(self, toolbar: str | QToolBar) -> QToolBar:
        """Add toolbar with specified name.

        :param toolbar: Name for the toolbar or QToolBar object.
        """
        if isinstance(toolbar, str):
            name = toolbar
            _toolbar = QToolBar(name, parent=self._mainWindow)
        else:
            name = toolbar.windowTitle()
            _toolbar = toolbar
        self._toolbars[name] = _toolbar
        return _toolbar

    def mapCanvas(self) -> QgsMapCanvas:
        """Return a pointer to the map canvas."""
        return self.canvas

    def mainWindow(self) -> QWidget:
        """Return a pointer to the main window.

        In case of QGIS it returns an instance of QgisApp.
        """
        return self._mainWindow

    def addDockWidget(self, area: int, dock_widget: QDockWidget) -> None:
        """Add a dock widget to the main window.

        :param area: Where in the ui the dock should be placed.
        :type area: Qt.DockWidgetArea

        :param dock_widget: A dock widget to add to the UI.
        :type dock_widget: QDockWidget
        """
        pass

    def removeDockWidget(self, dockwidget: QDockWidget) -> None:
        """Removes the specified dock widget from main window
        (without deleting it).
        """
        pass

    def legendInterface(self) -> QgsMapCanvas:
        """Get the legend."""
        return self.canvas

    def messageBar(self) -> MockMessageBar:
        """Get the messagebar"""
        return self._messageBar

    def getMockLayers(self) -> list[QgsMapLayer]:
        return self._layers

    def setActiveLayer(self, layer: QgsMapLayer) -> None:
        """Set the active layer (layer gets selected in the legend)"""
        self._active_layer_id = layer.id()

    def iconSize(self, dockedToolbar: bool) -> int:
        """Returns the toolbar icon size.
        :param dockedToolbar: If True, the icon size
        for toolbars contained within dockes is returned
        """
        if dockedToolbar:
            return QSize(16, 16)

        return QSize(24, 24)



#  Copyright (C) 2021-2026 pytest-qgis Contributors.
#
#
#  This file is part of pytest-qgis.
#
#  pytest-qgis is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  pytest-qgis is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with pytest-qgis.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import Any

from qgis.core import (
    QgsFeature,
    QgsFieldConstraints,
    QgsGeometry,
    QgsVectorDataProvider,
    QgsVectorLayer,
    QgsVectorLayerUtils,
)
from qgis.gui import QgisInterface, QgsAttributeDialog, QgsAttributeEditorContext
from qgis.PyQt.QtWidgets import QLabel, QWidget

from pytest_qgis import utils


class QgisBot:
    """
    Class to hold common utility methods for interacting with QIGS.
    """

    def __init__(  # noqa: QGS105 # Iface has to be passed in order to
        # ensure compatibility with all QGIS versions >= 3.10
        self,
        iface: QgisInterface,
    ) -> None:
        self._iface = iface

    def create_feature_with_attribute_dialog(  # noqa: PLR0913
        self,
        layer: QgsVectorLayer,
        geometry: QgsGeometry,
        attributes: dict[str, Any] | None = None,
        raise_from_warnings: bool = False,
        raise_from_errors: bool = True,
        show_dialog_timeout_milliseconds: int = 0,
    ) -> QgsFeature:
        """
        Create test feature with default values using QgsAttributeDialog.
        This ensures that all the default values are honored and
        for example boolean fields are either true or false, not null.

        :param layer: QgsVectorLayer to create feature into
        :param geometry: QgsGeometry of the feature
        :param attributes: attributes as a dictionary
        :param raise_from_warnings: Whether to raise error if there are non-enforcing
            constraint warnings with attribute values.
        :param raise_from_errors: Whether to raise error if there are enforcing
            constraint errors with attribute values.
        :param show_dialog_timeout_milliseconds: Shows attribute dialog. Useful for
            debugging.
        :return: Created QgsFeature that can be added to the layer.
        """

        initial_ids = set(layer.allFeatureIds())

        capabilities = layer.dataProvider().capabilities()

        if not capabilities & QgsVectorDataProvider.Capability.AddFeatures:
            raise ValueError(f"Could not create feature for the layer {layer.name()}")

        new_feature = QgsVectorLayerUtils.createFeature(
            layer, context=layer.createExpressionContext()
        )
        new_feature.setGeometry(geometry)

        if attributes is not None:
            if capabilities & QgsVectorDataProvider.Capability.ChangeAttributeValues:
                for field_name, value in attributes.items():
                    new_feature[field_name] = value
            else:
                raise ValueError(
                    f"Could not change attributes for layer {layer.name()}"
                )

        assert new_feature.isValid()

        warnings = {}
        errors = {}
        for field_index, field in enumerate(layer.fields()):
            no_warnings, warning_messages = QgsVectorLayerUtils.validateAttribute(
                layer,
                new_feature,
                field_index,
                QgsFieldConstraints.ConstraintStrength.ConstraintStrengthSoft,
            )
            no_errors, error_messages = QgsVectorLayerUtils.validateAttribute(
                layer,
                new_feature,
                field_index,
                QgsFieldConstraints.ConstraintStrength.ConstraintStrengthHard,
            )
            if not no_warnings:
                warnings[field.name()] = warning_messages
            if not no_errors:
                errors[field.name()] = error_messages

        if raise_from_warnings and warnings:
            raise ValueError(
                "There are non-enforcing constraint warnings in the attribute form: "
                f"{warnings!s}"
            )
        if raise_from_errors and errors:
            raise ValueError(
                "There are enforcing constraint errors in the attribute form: "
                f"{errors!s}"
            )

        context = QgsAttributeEditorContext()
        context.setMapCanvas(self._iface.mapCanvas())

        dialog = QgsAttributeDialog(
            layer, new_feature, False, self._iface.mainWindow(), True, context
        )
        dialog.show()
        dialog.setMode(QgsAttributeEditorContext.Mode.AddFeatureMode)

        # Process events to ensure all the signals are processed
        utils.wait(max(show_dialog_timeout_milliseconds, 1))

        # Two accepts to ignore warnings and errors
        dialog.accept()
        dialog.accept()

        feature_ids = set(layer.allFeatureIds())
        feature_id = list(feature_ids.difference(initial_ids))

        assert feature_id, "Creating new feature failed"
        return layer.getFeature(feature_id[0])

    @staticmethod
    def get_qgs_attribute_dialog_widgets_by_name(
        widget: QgsAttributeDialog | QWidget,
    ) -> dict[str, QWidget]:
        """
        Gets recursively all attribute dialog widgets by name.
        :param widget: QgsAttributeDialog for the first time, afterwards QWidget.
        :return: Dictionary with field names as keys and corresponding
        QWidgets as values.
        """
        widgets_by_name = {}
        for child in widget.children():
            if (
                isinstance(child, QLabel)
                and child.text() != ""
                and child.toolTip() != ""
                and child.buddy() is not None
            ):
                widgets_by_name[child.text()] = child.buddy()
            if hasattr(child, "children"):
                widgets_by_name = {
                    **widgets_by_name,
                    **QgisBot.get_qgs_attribute_dialog_widgets_by_name(child),
                }

        return widgets_by_name



#  Copyright (C) 2021-2026 pytest-qgis Contributors.
#
#
#  This file is part of pytest-qgis.
#
#  pytest-qgis is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  pytest-qgis is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with pytest-qgis.  If not, see <https://www.gnu.org/licenses/>.
#
import time
from collections import Counter
from collections.abc import Callable, Generator
from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any
from unittest.mock import MagicMock

from osgeo import gdal
from qgis.core import (
    QgsCoordinateReferenceSystem,
    QgsCoordinateTransform,
    QgsLayerTree,
    QgsLayerTreeGroup,
    QgsLayerTreeLayer,
    QgsMapLayer,
    QgsProcessing,
    QgsProject,
    QgsRasterLayer,
    QgsRectangle,
    QgsVectorLayer,
)
from qgis.PyQt import sip
from qgis.PyQt.QtCore import QCoreApplication

if TYPE_CHECKING:
    from _pytest.fixtures import FixtureRequest

DEFAULT_RASTER_FORMAT = "tif"

DEFAULT_EPSG = "EPSG:4326"
LAYER_KEYWORDS = ("layer", "lyr", "raster", "rast", "tif")


def get_common_extent_from_all_layers() -> QgsRectangle | None:
    """Get common extent from all QGIS layers in the project."""
    map_crs = QgsProject.instance().crs()
    layers = list(QgsProject.instance().mapLayers(validOnly=True).values())

    if layers:
        extent = transform_rectangle(layers[0].extent(), layers[0].crs(), map_crs)
        for layer in layers[1:]:
            extent.combineExtentWith(
                transform_rectangle(layer.extent(), layer.crs(), map_crs)
            )
        return extent
    return None


def set_map_crs_based_on_layers() -> None:
    """Set map crs based on layers of the project."""
    crs_counter = Counter(
        layer.crs().authid()
        for layer in QgsProject.instance().mapLayers().values()
        if layer.isSpatial()
    )
    if crs_counter:
        crs_id, _ = crs_counter.most_common(1)[0]
        crs = QgsCoordinateReferenceSystem(crs_id)
    else:
        crs = QgsCoordinateReferenceSystem(DEFAULT_EPSG)
    QgsProject.instance().setCrs(crs)


def transform_rectangle(
    rectangle: QgsRectangle,
    in_crs: QgsCoordinateReferenceSystem,
    out_crs: QgsCoordinateReferenceSystem,
) -> QgsRectangle:
    """
    Transform rectangle from one crs to other.
    """
    if in_crs == out_crs:
        return rectangle

    transform = QgsCoordinateTransform(
        QgsCoordinateReferenceSystem(in_crs),
        QgsCoordinateReferenceSystem(out_crs),
        QgsProject.instance(),
    )
    return transform.transformBoundingBox(rectangle)


def get_layers_with_different_crs() -> list[QgsMapLayer]:
    map_crs = QgsProject.instance().crs()
    return [
        layer
        for layer in QgsProject.instance().mapLayers().values()
        if layer.crs() != map_crs
    ]


def replace_layers_with_reprojected_clones(
    layers: list[QgsMapLayer], output_path: Path
) -> None:
    """
    For some reason all layers having differing crs from the project are invisible.
    Hotfix is to replace those by reprojected layers with map crs.
    """
    import processing  # noqa: PLC0415

    vector_layers = [
        layer
        for layer in layers
        if isinstance(layer, QgsVectorLayer) and layer.isSpatial()
    ]
    raster_layers = [
        layer
        for layer in layers
        if isinstance(layer, QgsRasterLayer) and layer.isSpatial()
    ]

    map_crs = QgsProject.instance().crs()
    for input_layer in vector_layers:
        output_layer: QgsVectorLayer = processing.run(  # noqa: QGS110
            "native:reprojectlayer",
            {
                "INPUT": input_layer,
                "TARGET_CRS": map_crs,
                "OUTPUT": QgsProcessing.TEMPORARY_OUTPUT,
            },
        )["OUTPUT"]
        if not output_layer.crs().isValid():
            output_layer.setCrs(map_crs)

        copy_layer_style_and_position(input_layer, output_layer, output_path)

    for input_layer in raster_layers:
        try:
            output_raster = str(
                Path(output_path, f"{input_layer.name()}.{DEFAULT_RASTER_FORMAT}")
            )
            warp = gdal.Warp(
                output_raster, input_layer.source(), dstSRS=map_crs.authid()
            )

        finally:
            warp = None  # noqa: F841

        output_layer = QgsRasterLayer(output_raster)
        if not output_layer.crs().isValid():
            output_layer.setCrs(map_crs)
        copy_layer_style_and_position(input_layer, output_layer, output_path)

    # Remove originals from project
    QgsProject.instance().removeMapLayers([layer.id() for layer in layers])


def copy_layer_style_and_position(
    layer1: QgsMapLayer, layer2: QgsMapLayer, tmp_path: Path
) -> None:
    """
    Copy layer style and position to another layer.
    """
    style_file = str(Path(tmp_path, f"{layer1.id()}.qml"))
    error_msg, succeeded = layer1.saveNamedStyle(style_file)
    if not succeeded:
        raise AssertionError(f"Failed to save layer style to {style_file}: {error_msg}")

    error_msg, succeeded = layer2.loadNamedStyle(style_file)
    if not succeeded:
        raise AssertionError(
            f"Failed to load layer style from {style_file}: {error_msg}"
        )
    layer2.setMetadata(layer1.metadata())
    layer2.setName(layer1.name())
    if layer2.isValid():  # noqa: SIM102
        if not QgsProject.instance().addMapLayer(layer2, False):
            raise AssertionError(f"Failed to add layer {layer2.name()} to project")

    root: QgsLayerTree = QgsProject.instance().layerTreeRoot()
    layer_tree_layer: QgsLayerTreeLayer = root.findLayer(layer1)
    group: QgsLayerTreeGroup = layer_tree_layer.parent()
    index = {child.name(): i for i, child in enumerate(group.children())}[
        layer_tree_layer.name()
    ]

    group.insertLayer(index + 1, layer2)


def clean_qgis_layer(fn: Callable[..., QgsMapLayer]) -> Callable[..., QgsMapLayer]:
    """
    Decorator to ensure that a map layer created by a fixture is cleaned properly.

    Sometimes fixture non-memory layers that are used but not added
    to the project might cause segmentation fault errors.

    This decorator works only with fixtures that **return** QgsMapLayer instances.
    There is no support for fixtures that use yield.

    >>> @pytest.fixture()
    >>> @clean_qgis_layer
    >>> def geojson_layer() -> QgsVectorLayer:
    >>>     layer = QgsVectorLayer("layer.json", "layer", "ogr")
    >>>     return layer

    This decorator is the alternative way of cleaning the layers since layer fixtures
    are automatically cleaned if they contain one of the keywords listed in
    LAYER_KEYWORDS by pytest_runtest_teardown hook.
    """

    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Generator[QgsMapLayer, None, None]:
        layer = fn(*args, **kwargs)
        yield layer
        _set_layer_owner_to_project(layer)

    return wrapper


def ensure_qgis_layer_fixtures_are_cleaned(request: "FixtureRequest") -> None:
    """
    Sometimes fixture non-memory layers that are used but not added
    to the project might cause segmentation fault errors.

    This function ensures that the layer fixtures will be cleaned by
    adding and removing those into the project.

    It does not matter what scoped the fixtures are since the
    layers are not actually deleted at any point.
    """
    for fixture_name in request.fixturenames:
        if any(
            possible_layer_name in fixture_name.lower()
            for possible_layer_name in LAYER_KEYWORDS
        ):
            try:
                layer = request.getfixturevalue(fixture_name)
            except AssertionError:
                continue
            _set_layer_owner_to_project(layer)


def _set_layer_owner_to_project(layer: Any) -> None:
    if (
        isinstance(layer, QgsMapLayer)
        and not isinstance(layer, MagicMock)
        and not sip.isdeleted(layer)
        and layer.id() not in QgsProject.instance().mapLayers(True)
    ):
        if not QgsProject.instance().addMapLayer(layer):
            raise AssertionError(f"Failed to add layer {layer.name()} to project")
        QgsProject.instance().removeMapLayer(layer)


def wait(wait_time_milliseconds: int = 0) -> None:
    """Waits for wait_time ms."""
    start = time.time()

    while (time.time() - start) * 1000 < wait_time_milliseconds:
        QCoreApplication.processEvents()



#  Copyright (C) 2021-2026 pytest-qgis Contributors.
#
#
#  This file is part of pytest-qgis.
#
#  pytest-qgis is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  pytest-qgis is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with pytest-qgis.  If not, see <https://www.gnu.org/licenses/>.
import typing

from qgis.core import Qgis
from qgis.gui import QgsMessageBar as QgsMessageBarOriginal


class MockMessageBar(QgsMessageBarOriginal):
    """Mocked message bar to hold the messages."""

    def __init__(self) -> None:
        super().__init__(None)
        self.messages: dict[int, list[str]] = {}
        self._init_messages()

    def _init_messages(self) -> None:
        self.messages = {
            Qgis.MessageLevel.Info: [],
            Qgis.MessageLevel.Warning: [],
            Qgis.MessageLevel.Critical: [],
            Qgis.MessageLevel.Success: [],
        }

    def get_messages(self, level: int) -> list[str]:
        """Used to test which messages have been logged."""
        return self.messages[level]

    def clear_messages(self) -> None:
        """Clear logged messages."""
        self._init_messages()

    @typing.overload
    def pushMessage(
        self,
        text: str | None = None,
        level: Qgis.MessageLevel = Qgis.MessageLevel.Info,
        duration: int = -1,
    ) -> None: ...

    @typing.overload
    def pushMessage(
        self,
        title: str | None = None,
        text: str | None = None,
        level: Qgis.MessageLevel = Qgis.MessageLevel.Info,
        duration: int = -1,
    ) -> None: ...

    @typing.overload
    def pushMessage(
        self,
        title: str | None = None,
        text: str | None = None,
        showMore: str | None = None,  # noqa: N803
        level: Qgis.MessageLevel = Qgis.MessageLevel.Info,
        duration: int = -1,
    ) -> None: ...

    @typing.no_type_check
    def pushMessage(  # noqa: N802
        self,
        *args: str | int,
        **kwargs: dict[str, str | int],
    ) -> None:
        """A mocked method for pushing a message to the bar."""
        title = kwargs.get("title")
        text = kwargs.get("text")
        level = kwargs.get("level", Qgis.MessageLevel.Info)

        length = len(args)
        if length == 1 and not text:
            text = args[0]
        elif length == 1 and not title:
            title = args[0]
        elif length > 1 and isinstance(args[1], str):
            # title, text, level, ...
            title = args[0]
            text = args[1]
            if length > 2 and isinstance(args[2], int):  # noqa: PLR2004
                level = args[2]
            elif length > 3 and isinstance(args[3], int):  # noqa: PLR2004
                level = args[3]
        elif length > 1 and isinstance(args[1], int):
            # text, level, ...
            text = args[0]
            level = args[1]
        elif args and not text:
            text = ", ".join(map(str, args))

        msg = f"{title or 'no-title'}:{text}"
        self.messages[level].append(msg)

    # NOTE: should be a slot
    def pushWarning(  # noqa: N802
        self,
        title: str,
        message: str,
    ) -> None:
        """Pushes a warning message that must be manually dismissed by the user."""
        self.pushMessage(title, message, Qgis.MessageLevel.Warning, -1)
